package com.bureau.onetaplogin

import android.content.Context
import android.net.ConnectivityManager
import android.net.ConnectivityManager.NetworkCallback
import android.net.Network
import android.net.NetworkCapabilities
import android.net.NetworkRequest
import android.os.Build
import android.util.Log
import androidx.annotation.Keep
import com.bureau.base.Environment
import com.bureau.onetaplogin.data.remote.RemoteDataSourceImpl
import com.bureau.onetaplogin.exceptions.AuthenticationException
import com.bureau.onetaplogin.models.AuthCallback
import com.bureau.onetaplogin.models.AuthenticationStatus
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.delay
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import java.io.IOException

@Keep
class BureauAuth private constructor(
    private val environment: Environment,
    private val clientId: String,
    private val timeoutInMs: Int
) {
    private val authenticationStatusFlow = MutableStateFlow(AuthenticationStatus.UnknownState)
    private val coroutineScope = CoroutineScope(Dispatchers.IO)
    private var resultSent = false

    fun authenticate(
        context: Context,
        correlationId: String,
        mobileNumber: Long,
        authCallback: AuthCallback
    ) {
        resultSent = false
        coroutineScope.launch {
            delay(timeoutInMs.toLong())
            if (authenticationStatusFlow.value == AuthenticationStatus.UnknownState && !resultSent) {
                authCallback.onResult(AuthenticationStatus.UnknownState)
                resultSent = true
            }
        }

        coroutineScope.launch {
            authenticationStatusFlow.collect {
                if (it != AuthenticationStatus.UnknownState && !resultSent) {
                    authCallback.onResult(it)
                    resultSent = true
                }
            }
        }

        // some devices do not support request network api call. To mitigate we first do a direct auth and then retry via RequestNetwork api
        val status =
            triggerAuthenticationFlowDirect(context, correlationId, mobileNumber)
        if (!status) {
            triggerAuthenticationFlowViaConnectivityManager(
                context,
                correlationId,
                mobileNumber
            )
        }
    }

    private fun triggerAuthenticationFlowDirect(
        context: Context,
        correlationId: String,
        mobileNumber: Long
    ): Boolean {
        val connectivityManager =
            context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager

        // call auth directly if cellular network is
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            val activeNetworkInfo = connectivityManager.activeNetwork
            if (activeNetworkInfo != null) {
                val networkCapabilities =
                    connectivityManager.getNetworkCapabilities(activeNetworkInfo)
                if (networkCapabilities != null && networkCapabilities.hasTransport(
                        NetworkCapabilities.TRANSPORT_CELLULAR
                    )
                ) {
                    Log.i("BureauAuth", "Android M trigger Auth")
                    triggerAuthenticationFlow(
                        correlationId,
                        mobileNumber
                    )
                    return true
                }
            }
        }
        return false
    }

    private fun triggerAuthenticationFlowViaConnectivityManager(
        context: Context,
        correlationId: String,
        mobileNumber: Long
    ) {
        val connectivityManager =
            context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
        val networkRequest = NetworkRequest.Builder()
            .addTransportType(NetworkCapabilities.TRANSPORT_CELLULAR)
            .addCapability(NetworkCapabilities.NET_CAPABILITY_INTERNET)
            .build()
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            connectivityManager.requestNetwork(
                networkRequest,
                registerNetworkCallbackForOPlusDevices(
                    correlationId,
                    mobileNumber,
                    connectivityManager
                ),
                timeoutInMs
            )
        } else if (Build.VERSION.SDK_INT == Build.VERSION_CODES.M) {
            // https://stackoverflow.com/questions/32185628/connectivitymanager-requestnetwork-in-android-6-0
            // Android M has a bug where requestNetwork never works, hence instead of requestNetwork,
            // call api directly if on cellular network else fail the authenticate request
            Log.e("BureauAuth", "Android M No network")
            authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.NetworkUnavailable)
        } else {
            connectivityManager.requestNetwork(
                networkRequest,
                registerCallbackForOMinusDevices(
                    correlationId,
                    mobileNumber
                )
            )
        }
    }

    private fun registerCallbackForOMinusDevices(
        correlationId: String,
        mobileNumber: Long
    ): NetworkCallback {
        return object : NetworkCallback() {
            override fun onUnavailable() {
                super.onUnavailable()
                authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.NetworkUnavailable)
            }

            override fun onAvailable(network: Network) {
                super.onAvailable(network)
                triggerAuthenticationFlow(
                    correlationId,
                    mobileNumber,
                    network
                )
            }
        }
    }

    private fun registerNetworkCallbackForOPlusDevices(
        correlationId: String,
        mobileNumber: Long,
        connectivityManager: ConnectivityManager
    ): NetworkCallback {
        return object : NetworkCallback() {
            override fun onUnavailable() {
                super.onUnavailable()
                val activeNetworkInfo = connectivityManager.activeNetworkInfo
                if (activeNetworkInfo != null && activeNetworkInfo.isConnected) {
                    authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.OnDifferentNetwork)
                } else {
                    authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.NetworkUnavailable)
                }
            }

            override fun onAvailable(network: Network) {
                super.onAvailable(network)
                triggerAuthenticationFlow(
                    correlationId,
                    mobileNumber,
                    network
                )
            }
        }
    }

    private fun triggerAuthenticationFlow(
        correlationId: String,
        mobileNumber: Long,
        network: Network? = null
    ) {
        coroutineScope.launch {
            try {
                val remoteDataSource = RemoteDataSourceImpl(environment, timeoutInMs, network)
                val response = remoteDataSource.makeInitiateCall(
                    clientId,
                    correlationId,
                    mobileNumber.toString()
                )
                when {
                    response.code() == 200 -> {
                        authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.Completed)
                        Log.i("BureauAuth", "Auth Success")
                    }
                    response.code() == 401 -> {
                        authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.ExceptionOnAuthenticate)
                        Log.i("BureauAuth", "Auth failed due to Authentication Exception")
                    }
                    response.code() == 500 -> {
                        Log.i("BureauAuth", "Auth failed due to ServerError")
                        authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.UnknownState)
                    }
                    response.code() == 400 -> {
                        Log.i("BureauAuth", "Auth failed due to possible wrong network")
                        authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.OnDifferentNetwork)
                    }
                    response.code() == 422 -> {
                        Log.i("BureauAuth", "Auth failed due to possible credential expiry")
                        authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.UnknownState)
                    }
                    else -> {
                        Log.i("BureauAuth", "Auth failed due to unknown error")
                        authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.UnknownState)
                    }
                }
            } catch (e: AuthenticationException) {
                Log.i("BureauAuth", e.message)
                authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.ExceptionOnAuthenticate)
            } catch (e: IOException) {
                Log.i("BureauAuth", e.message ?: "")
                authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.ExceptionOnAuthenticate)
            } catch (e: RuntimeException) {
                Log.i("BureauAuth", e.message ?: "")
                authenticationStatusFlow.compareAndSet(AuthenticationStatus.UnknownState, AuthenticationStatus.ExceptionOnAuthenticate)
            }
        }
    }

    @Keep
    data class Builder(
        private var environment: Environment = Environment.ENV_PRODUCTION,
        private var clientId: String = "",
        private var timeOutInMs: Int = 15 * 1000
    ) {
        fun environment(environment: Environment) = apply {
            this.environment = environment
        }

        fun clientId(clientId: String) = apply {
            this.clientId = clientId
        }

        fun timeOutInMs(timeOutInMs: Int) = apply {
            this.timeOutInMs = timeOutInMs
        }

        fun build(): BureauAuth {
            return BureauAuth(environment, clientId, timeOutInMs)
        }
    }
}
